/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2018 Red Hat, Inc., and individual contributors
 * as indicated by the @author tags.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.jboss.threads;

/**
 * EQE base class: padding.
 */
abstract class EnhancedQueueExecutorBase4 extends EnhancedQueueExecutorBase3 {
    /**
     * Padding fields.
     */
    @SuppressWarnings("unused")
    int p00, p01, p02, p03,
        p04, p05, p06, p07,
        p08, p09, p0A, p0B,
        p0C, p0D, p0E, p0F;

    EnhancedQueueExecutorBase4() {}
}
